@@ -309,6 +309,19 @@ static void service_number_free(struct service_number *num)
 	g_free(num);
 }
 
+static void call_state_watches(struct ofono_sim *sim)
+{
+	GSList *l;
+	ofono_sim_state_event_cb_t notify;
+
+	for (l = sim->state_watches->items; l; l = l->next) {
+		struct ofono_watchlist_item *item = l->data;
+		notify = item->notify;
+
+		notify(sim->state, item->notify_data);
+	}
+}
+
 static DBusMessage *sim_get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -806,6 +819,9 @@ static DBusMessage *sim_enter_pin(DBusConnection *conn, DBusMessage *msg,
 	if (type == OFONO_SIM_PASSWORD_NONE || type != sim->pin_type)
 		return __ofono_error_invalid_format(msg);
 
+	if (password_is_pin(type) == FALSE)
+		return __ofono_error_invalid_format(msg);
+
 	if (!__ofono_is_valid_sim_pin(pin, type))
 		return __ofono_error_invalid_format(msg);
 
@@ -1355,25 +1371,18 @@ static void sim_ready(enum ofono_sim_state new_state, void *user)
 
 static void sim_set_ready(struct ofono_sim *sim)
 {
-	GSList *l;
-	ofono_sim_state_event_cb_t notify;
-
 	if (sim == NULL)
 		return;
 
-	if (sim->state != OFONO_SIM_STATE_INSERTED)
+	if (sim->state != OFONO_SIM_STATE_INSERTED &&
+			sim->state != OFONO_SIM_STATE_LOCKED_OUT)
 		return;
 
 	sim->state = OFONO_SIM_STATE_READY;
 
 	sim_fs_check_version(sim->simfs);
 
-	for (l = sim->state_watches->items; l; l = l->next) {
-		struct ofono_watchlist_item *item = l->data;
-		notify = item->notify;
-
-		notify(sim->state, item->notify_data);
-	}
+	call_state_watches(sim);
 }
 
 static void sim_imsi_cb(const struct ofono_error *error, const char *imsi,
@@ -2280,9 +2289,6 @@ static void sim_free_state(struct ofono_sim *sim)
 
 void ofono_sim_inserted_notify(struct ofono_sim *sim, ofono_bool_t inserted)
 {
-	ofono_sim_state_event_cb_t notify;
-	GSList *l;
-
 	if (inserted == TRUE && sim->state == OFONO_SIM_STATE_NOT_PRESENT)
 		sim->state = OFONO_SIM_STATE_INSERTED;
 	else if (inserted == FALSE && sim->state != OFONO_SIM_STATE_NOT_PRESENT)
@@ -2294,13 +2300,7 @@ void ofono_sim_inserted_notify(struct ofono_sim *sim, ofono_bool_t inserted)
 		return;
 
 	sim_inserted_update(sim);
-
-	for (l = sim->state_watches->items; l; l = l->next) {
-		struct ofono_watchlist_item *item = l->data;
-		notify = item->notify;
-
-		notify(sim->state, item->notify_data);
-	}
+	call_state_watches(sim);
 
 	if (inserted)
 		sim_initialize(sim);
@@ -2386,9 +2386,8 @@ static void sim_pin_query_cb(const struct ofono_error *error,
 			/* Force the sim state out of READY */
 			sim_free_main_state(sim);
 
-			sim->state = OFONO_SIM_STATE_INSERTED;
-			__ofono_modem_sim_reset(
-					__ofono_atom_get_modem(sim->atom));
+			sim->state = OFONO_SIM_STATE_LOCKED_OUT;
+			call_state_watches(sim);
 		}
 		break;
 	}
